/*
 * Copyright 2010 IHE International (http://www.ihe.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.ihe.gazelle.proxy.admin.gui;

import java.io.Serializable;

import javax.ejb.Remove;

import net.ihe.gazelle.common.interfacegenerator.GenerateInterface;
import net.ihe.gazelle.proxy.admin.model.ApplicationConfiguration;

import org.jboss.seam.Component;
import org.jboss.seam.ScopeType;
import org.jboss.seam.annotations.AutoCreate;
import org.jboss.seam.annotations.Destroy;
import org.jboss.seam.annotations.Name;
import org.jboss.seam.annotations.Scope;
import org.jboss.seam.security.Identity;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * <b>Class Description : </b>ApplicationConfigurationManager<br>
 * <br>
 * This class contains the methods which are called in the footer modal panels
 * 
 * @class ApplicationConfigurationManager.java
 * @package net.ihe.gazelle.simulator.common.action
 * @author Anne-Gaëlle Bergé / INRIA Rennes IHE development Project
 * @see > Anne-Gaelle.Berge@inria.fr - http://www.ihe-europe.org
 * @version 1.0 - 2010, October 15th
 * 
 */

@Name("applicationConfigurationManager")
@AutoCreate
@Scope(ScopeType.APPLICATION)
@GenerateInterface("ApplicationConfigurationManagerLocal")
public class ApplicationConfigurationManager implements Serializable,
		ApplicationConfigurationManagerLocal {

	/**
	 * 
	 */
	private static final long serialVersionUID = -7817765881624341837L;
	private static Logger log = LoggerFactory
			.getLogger(ApplicationConfigurationManager.class);

	private String releaseNoteUrl;
	private String documentation;
	private Boolean worksWithoutCas;
	private String issueTracker;
	private String evsClientUrl;
	private String dicomStorage;
	private String proxyIpAddresses;
	private Integer minProxyPort;
	private Integer maxProxyPort;
	private String applicationUrl;
	private String casUrl;
	private String proxyOid;
	private Boolean ipLogin;
	private String ipLoginAdmin;

	@Remove
	@Destroy
	public void destroy() {
		resetApplicationConfiguration();
	}

	public void resetApplicationConfiguration() {
		releaseNoteUrl = null;
		documentation = null;
		worksWithoutCas = null;
		issueTracker = null;
		evsClientUrl = null;
		dicomStorage = null;
		proxyIpAddresses = null;
		maxProxyPort = null;
		minProxyPort = null;
		casUrl = null;
		applicationUrl = null;
		proxyOid = null;
		ipLogin = null;
		ipLoginAdmin = null;
	}

	private String getApplicationProperty(String inPropertyKey) {
		String value = ApplicationConfiguration
				.getValueOfVariable(inPropertyKey);
		if (value == null) {
			log.warn(inPropertyKey
					+ " variable is missing in app_configuration table");
		}
		return value;
	}

	public String getReleaseNoteUrl() {
		if (releaseNoteUrl == null) {
			releaseNoteUrl = getApplicationProperty("application_release_notes_url");
		}
		return releaseNoteUrl;
	}

	public String getIssueTracker() {
		if (issueTracker == null) {
			issueTracker = getApplicationProperty("application_issue_tracker");
		}
		return issueTracker;
	}

	public String getDocumentation() {
		if (this.documentation == null) {
			this.documentation = getApplicationProperty("application_documentation");
		}
		if (this.documentation == null) {
			this.documentation = "http://gazelle.ihe.net";
		}
		return this.documentation;
	}

	public boolean isWorksWithoutCas() {
		if (worksWithoutCas == null) {
			String booleanAsString = getApplicationProperty("application_works_without_cas");
			if (booleanAsString == null) {
				this.worksWithoutCas = false;
			} else {
				this.worksWithoutCas = Boolean.valueOf(booleanAsString);
			}
		}
		return worksWithoutCas;
	}

	/**
	 * Returns true if the logged in user has role "admin_role" or if the
	 * application works without cas
	 * 
	 * @return
	 */
	public boolean isUserAllowedAsAdmin() {
		return Identity.instance().isLoggedIn()
				&& Identity.instance().hasRole("admin_role");
	}

	public static ApplicationConfigurationManager instance() {
		return (ApplicationConfigurationManager) Component
				.getInstance("applicationConfigurationManager");
	}

	public String getEvsClientUrl() {
		if (evsClientUrl == null) {
			evsClientUrl = getApplicationProperty("evs_client_url");
		}
		return evsClientUrl;
	}

	public String getDicomStorage() {
		if (dicomStorage == null) {
			dicomStorage = getApplicationProperty("storage_dicom");
		}
		return dicomStorage;
	}

	public String getProxyIpAddresses() {
		if (proxyIpAddresses == null) {
			proxyIpAddresses = getApplicationProperty("proxy_ip_addresses");
		}
		return proxyIpAddresses;
	}

	@Override
	public int getMinProxyPort() {
		if (minProxyPort == null) {
			String port = getApplicationProperty("min_proxy_port");
			if (port != null) {
				minProxyPort = Integer.decode(port);
			} else {
				minProxyPort = 10101;
			}
		}
		return minProxyPort;
	}

	public int getMaxProxyPort() {
		if (maxProxyPort == null) {
			String port = getApplicationProperty("max_proxy_port");
			if (port != null) {
				maxProxyPort = Integer.decode(port);
			} else {
				maxProxyPort = 11000;
			}
		}
		return maxProxyPort;
	}

	public String getApplicationUrl() {
		if (applicationUrl == null) {
			applicationUrl = getApplicationProperty("application_url");
		}
		return applicationUrl;
	}

	public String getCasUrl() {
		if (casUrl == null) {
			casUrl = getApplicationProperty("cas_url");
		}
		return casUrl;
	}

	public String getProxyOid() {
		if (proxyOid == null) {
			proxyOid = getApplicationProperty("proxy_oid");
		}
		return proxyOid;
	}

	public Boolean getIpLogin() {
		if (ipLogin == null) {
			String booleanAsString = ApplicationConfiguration
					.getValueOfVariable("ip_login");
			if (booleanAsString == null) {
				this.ipLogin = false;
			} else {
				this.ipLogin = Boolean.valueOf(booleanAsString);
			}
		}
		return ipLogin;
	}

	public String getIpLoginAdmin() {
		if (this.ipLoginAdmin == null) {
			this.ipLoginAdmin = ApplicationConfiguration
					.getValueOfVariable("ip_login_admin");
		}
		return ipLoginAdmin;
	}

	public String loginByIP() {
		log.info("starting authentication without cas");
		Identity identity = (Identity) Component
				.getInstance("org.jboss.seam.security.identity");
		if ("loggedIn".equals(identity.login())) {
			return "/home.xhtml";
		} else {
			return null;
		}
	}
}
